#-------------------------------------------
#
# Copyright (c) 2001 DataChannel, Inc. All Rights Reserved.
#
# LGPL 2.1
#
# $Id:$
#
#-------------------------------------------

from __main__ import __builtins__
import sys
import thread
import ZODB.Transaction
from zLOG import LOG, TRACE, DEBUG, BLATHER, INFO, PROBLEM, ERROR, PANIC

_idToTransactionMap = {}
_idToTransactionMapLock = thread.allocate_lock()

_autoCreateTransactions = 1
def autoCreateTransactions(onoff):
    global _autoCreateTransactions
    _autoCreateTransactions = onoff

def get_transaction(_id=thread.get_ident):
    id = _id()
    result = _idToTransactionMap.get(id, None)
    if not result:
        if _autoCreateTransactions:
            result = Transaction(id)
            print "autocreating transaction:%s" % result
            result.attach(_id)
        else:
            raise KeyError(id)
    return result

# Replace global definition of get_transaction
__builtins__.get_transaction = get_transaction

class Transaction(ZODB.Transaction.Transaction):
    def __init__(self, id=None):
        ZODB.Transaction.Transaction.__init__(self, id)
        self._onCommitCallbacks = None

    def attach(self, _id=thread.get_ident, force=0):
        id = _id()
        _idToTransactionMapLock.acquire()
        try:
            if not _idToTransactionMap.has_key(id) or force:
                _idToTransactionMap[id] = self
            else:
                raise KeyError("%s already register to %s" % \
                               (id, _idToTransactionMap.get(id, None)))
        finally:
            _idToTransactionMapLock.release()

    def detach(self, _id=thread.get_ident):
        id = _id()
        try:
            del _idToTransactionMap[id]
        finally:
            pass

    def registerCommitCallback(self, func, *args, **kw):
        if self._onCommitCallbacks is None:
            self._onCommitCallbacks = []
        self._onCommitCallbacks.append((func, args, kw))

    def abort(self, subtransaction=0, freeme=1):
        ZODB.Transaction.Transaction.abort(self, subtransaction, freeme)
        self._onCommitCallbacks = None
        
    def commit(self, subtransaction=None):
        ZODB.Transaction.Transaction.commit(self, subtransaction)
        if subtransaction is None and self._onCommitCallbacks is not None:
            for callback, args, kw in self._onCommitCallbacks:
                try:
                    apply(callback, args, kw)
                except:
                    LOG('ThorTransaction',ERROR, 'Error processing callbacks',
                        error=sys.exc_info())
            self._onCommitCallbacks = None

class CommitActivity:
    """Mix-in class that provides transaction management support

    A sub class should call self._register() whenever it performs
    any transaction-dependent operations (e.g. sql statements).

    The sub class will need to override:
      begin if necessary
      vote to raise an except if necessary
      finish to finallize work,
      abort to roll-back work
    """
    def __init__(self):
        self._beginCalled = 0
        self._voteCalled = 0
        self._abortCalled = 0
        self._finishCalled = 0
        self._register()

    _registered=None

    def _begin(self):
        """Hook method to begin external transaction.
        This may be called multiple times,"""
        pass

    def _vote(self):
        """Hook method to vote on success of transaction commit.
        This will only be called once,"""
        pass

    def _finish(self):
        """Hook method to complete transaction work.
        This may be called multiple times."""
        pass

    def _abort(self):
        """Hook method to undo transaction work.
        This may be called multiple times."""
        pass

    def _postCommit(self):
        """Hook method called when Transaction has completed a
        commit process.
        """

    class Surrogate:
        "A ZODB persistent look-alike."
        def __init__(self, connection):
            self._p_jar=connection

    def _register(self):
        """This method will register a surrogate persistent
        object with the transaction.  This surrogate will include
        me in the tpc_* two-phase commit process."""

        if not self._registered:
            try:
                trans = get_transaction()
                trans.register(CommitActivity.Surrogate(self))
                trans.registerCommitCallback(self._postCommit)
                self._registered=1
            except:
                import traceback
                traceback.print_exc()
                pass

    def tpc_begin(self, transaction, sub=None):
        "Connection method to begin a commit."
        if sub is None and not self._beginCalled:
            self._beginCalled = 1
            self._begin()

    def tpc_vote(self, *ignored):
        "Connection method to vote on commit success."
        if self._beginCalled and not self._voteCalled:
            self._voteCalled = 1
            self._vote()
        
    def tpc_finish(self, *ignored):
        """Connection method to signal commit success.
        This should *never* fail"""
        if self._beginCalled and not self._finishCalled:
            self._finishCalled = 1
            try:
                self._finish()
            finally:
                self._registered=0

    def tpc_abort(self, *ignored):
        """Connnection method to signal commit failure.
        This should *never* fail"""
        if self._beginCalled and not self._abortCalled:
            self._abortCalled = 1
            try:
                self._abort()
            finally:
                self._registered=0

    def commit(self, *ignored):
        "We don't need to actually do any storing."
        pass

    def abort(self, *ignored):
        "We don't need to actually do any aborting."
        pass

if __name__ == '__main__':
    t = get_transaction()

    assert thread.get_ident() in t.attachedTo(), \
           "Should include this thread id"
    
    assert t == get_transaction()
    
    t.detach()

    assert t != get_transaction()

    autoCreateTransactions(0)

    get_transaction().detach()

    try:
        get_transaction()
    except KeyError:
        pass
    else:
        assert 0, 'Shouldn\'t have gotten a transaction'

    t1 = Transaction()
    t2 = Transaction()
    t1.attach()
    try:
        t2.attach()
    except KeyError:
        pass
    else:
        assert 0, "Should have failed because of t1"
